/* PConnection_libusb.c
 *
 * USB-related stuff for Visors.
 *
 *	Copyright (C) 2000, Louis A. Mamakos.
 *	Copyright (C) 2002, Alessandro Zummo.
 *	You may distribute this file under the terms of the Artistic
 *	License, as specified in the README file.
 *
 * $Id$
 */

#include "config.h"

#include <stdio.h>              /* This is left outside of the #if WITH_USB solely
                                 * to make 'gcc' shut up: apparently, ANSI C
                                 * doesn't like empty source files.
                                 */

#if WITH_LIBUSB                 /* This encompasses the entire file */
#  include <unistd.h>
#  include <stdlib.h>
#  include <fcntl.h>
#  include <errno.h>

#  if STDC_HEADERS
#    include <string.h>         /* For memcpy() et al. */
#  else                         /* STDC_HEADERS */
#    ifndef HAVE_STRCHR
#      define strchr index
#      define strrchr rindex
#    endif                      /* HAVE_STRCHR */
#    ifndef HAVE_MEMCPY
#      define memcpy(d,s,n)		bcopy ((s), (d), (n))
#      define memmove(d,s,n)	bcopy ((s), (d), (n))
#    endif                      /* HAVE_MEMCPY */
#  endif                        /* STDC_HEADERS */

#  include <usb.h>

#  if HAVE_LIBINTL_H
#    include <libintl.h>        /* For i18n */
#  endif                        /* HAVE_LIBINTL_H */

#  include "pconn/PConnection.h"
#  include "palm.h"
#  include "pconn/palm_errno.h"
#  include "pconn/netsync.h"

#  define IOBUF_LEN	4096

struct usb_data {
  usb_dev_handle *dev;
  unsigned char hotsync_ep_in;
  unsigned char hotsync_ep_out;
  unsigned char iobuf[IOBUF_LEN];
  unsigned char *iobufp;
  int iobuflen;
};

#  include <linux/types.h>
#  include "visor.h"

static char *hs_usb_functions[] = {
  "Generic",
  "Debugger",
  "Hotsync",
  "Console",
  "RemoteFileSys",
  NULL
};

/*************************************************************************/

static int
libusb_bind(PConnection * pconn, const void *addr, const int addrlen)
{
  switch (pconn->protocol) {
  case PCONN_STACK_FULL:
    return slp_bind(pconn, (const struct slp_addr *) addr);

  case PCONN_STACK_SIMPLE:     /* Fall through */
  case PCONN_STACK_NET:
    return 0;

  case PCONN_STACK_NONE:
  case PCONN_STACK_DEFAULT:
    /* XXX - Error */
  default:
    /* XXX - Indicate error: unsupported protocol */
    return -1;
  }
}

static int libusb_read(PConnection * p, unsigned char *buf, int len)
{
  /*
   *  We've got to do intermediate buffering of the USB data
   *  from the Visor's bulk endpoint.  This is because the kernel
   *  USB driver isn't buffering any data, and we must have a read
   *  operation pending at least as large as the USB transfer size
   *  might potentially be.  
   *
   *  Here, we use a 1024 byte buffer, and return data out of
   *  it.  This could return "short" reads to the caller, but
   *  isn't semantically any different than reading an TTY device
   *  which won't return a necessarily predicatable amount of
   *  data.
   */

  struct usb_data *u = p->io_private;
  int copy_len, retlen = 0, readlen;

  do {
    /* (?) Return leftover stuff from the previous read.
     */
    if (u->iobuflen > 0) {
      copy_len = (len > u->iobuflen) ? u->iobuflen : len;

      memcpy(buf, u->iobufp, copy_len);
      /* XXX - Potential buffer overflow? */
      u->iobufp += copy_len;
      u->iobuflen -= copy_len;
      buf += copy_len;
      len -= copy_len;
      retlen += copy_len;
    }

    if (retlen == 0) {
      /* (?) There wasn't anything left over from the
       * last read(). Read some new stuff.
       */
      if (u->iobuflen > 0) {
        fprintf(stderr,
                _("libusb: trying to fill a non-empty " "buffer.\n"));
        abort();
      }

      u->iobufp = u->iobuf;
      if (len > sizeof(u->iobuf)) {
        readlen = sizeof(u->iobuf);
      } else {
        readlen = len;
      }

      if (readlen < 64)
        readlen = 64;

      IO_TRACE(2)
          fprintf(stderr, "calling usb_bulk_read(%02x)\n",
                  u->hotsync_ep_out | 0x80);
      u->iobuflen =
          usb_bulk_read(u->dev, u->hotsync_ep_out | 0x80,
                        (char *) u->iobufp, readlen, 5000);
      IO_TRACE(2)
          fprintf(stderr, "req %d,usb read %d, ret %d\n", len, readlen,
                  u->iobuflen);
      if (u->iobuflen < 0) {
        fprintf(stderr, "req %d,usb read: %s\n", len, usb_strerror());
        return u->iobuflen;
/*
        u->iobuflen = usb_bulk_read(u->dev, u->hotsync_ep_out | 0x80,
                                    (char *) u->iobufp, readlen, 500);

        if (u->iobuflen < 0) {
          fprintf(stderr, "usb read2: %s\n", usb_strerror());
        }
*/
      }
    }
  } while (retlen == 0);

  if (retlen > 0)
    p->bytes_read += retlen;

  return retlen;
}

static int libusb_write(PConnection * p, unsigned const char *buf,
                        const int len)
{
  struct usb_data *u = p->io_private;
  int ret;

  ret = usb_bulk_write(u->dev, u->hotsync_ep_in, (char *) buf, len, 5000);
  IO_TRACE(2)
      fprintf(stderr, "usb write %d, ret %d\n", len, ret);
  if (ret < 0)
    fprintf(stderr, "usb write: %s\n", usb_strerror());

  if (ret > 0)
    p->bytes_write += ret;

  return ret;
}

static int libusb_connect(PConnection * p, const void *addr,
                          const int addrlen)
{
  return -1;                    /* Not applicable to USB connection */
}

static int libusb_accept(PConnection * pconn)
{
  int err;
  udword newspeed;              /* Not really a speed; this is just
                                 * used for the return value from
                                 * cmp_accept().
                                 */

  switch (pconn->protocol) {
  case PCONN_STACK_FULL:
    /* Negotiate a CMP connection. Since this is USB, the speed
     * is meaningless: it'll just go however fast it can.
     */
    newspeed = cmp_accept(pconn, 0);
    if (newspeed == ~0)
      return -1;
    break;

  case PCONN_STACK_SIMPLE:     /* Fall through */
  case PCONN_STACK_NET:
    /* XXX - The m500 hangs right around here. The sync hangs
     * until the Palm aborts. This suggests that the Palm is
     * waiting for data from the desktop. Or maybe the Palm
     * wants to be tickled in some non-obvious way.
     */
    IO_TRACE(5)
        fprintf(stderr, "libusb_accept simple/net\n");

    /* XXX - It seems to help here to send a packet to the m500. */
    err = ritual_exch_server(pconn);
    if (err < 0) {
      IO_TRACE(3)
          fprintf(stderr, "libusb_accept simple/net: "
                  "ritual_exch_server() returned %d\n", err);
      return -1;
    }
    break;

  case PCONN_STACK_NONE:
  case PCONN_STACK_DEFAULT:
    /* XXX - Error */
  default:
    /* XXX - Indicate error: unsupported protocol */
    return -1;
  }

  return 0;
}

static int pconn_libusb_close(PConnection * p)
{
  struct usb_data *u = p->io_private;

  /* I _think_ it's okay to assume that pconn->protocol has been set,
   * even though this function may have been called before the
   * connection was completely set up.
   */

  /* Clean up the protocol stack elements */
  switch (p->protocol) {
  case PCONN_STACK_DEFAULT:    /* Fall through */
  case PCONN_STACK_FULL:
    dlp_tini(p);
    padp_tini(p);
    slp_tini(p);
    break;

  case PCONN_STACK_SIMPLE:     /* Fall through */
  case PCONN_STACK_NET:
    dlp_tini(p);
    netsync_tini(p);
    break;

  default:
    /* Do nothing silently: the connection might not have been
     * completely set up.
     */
    break;
  }

  if (u->dev)
    usb_close(u->dev);

  if (u != NULL)
    free((void *) u);

  return 0;
}

static int libusb_select(PConnection * p, pconn_direction which,
                         struct timeval *tvp)
{
  return 1;
}

static int libusb_drain(PConnection * p)
{
  /* We don't check p->protocol, because there's nothing to do for
   * any of them.
   */
  return 0;
}

static int get_hotsync_ep_ext(struct usb_data *u)
{
  int i, ret;
  unsigned char in, out;
  struct palm_ext_connection_info ci_ext;

  bzero((void *) &ci_ext, sizeof(ci_ext));

  ret =
      usb_control_msg(u->dev, USB_TYPE_VENDOR | USB_RECIP_ENDPOINT | 0x80,
                      PALM_GET_EXT_CONNECTION_INFORMATION, 0, 0,
                      (char *) &ci_ext, 0x14, 500);

  if (ret < 0) {
    perror(_
           ("usb_control_msg(PALM_GET_EXT_CONNECTION_INFORMATION) 1 failed"));
/*
	ret = usb_control_msg(u->dev, USB_TYPE_VENDOR | USB_RECIP_ENDPOINT | 0x80,
		PALM_GET_EXT_CONNECTION_INFORMATION, 0, 0, (char *)&ci_ext, 0x14, 500);

	if (ret<0) perror(_("usb_control_msg(PALM_GET_EXT_CONNECTION_INFORMATION) 1 failed"));
*/
  }

  if (ret > 0) {
    for (i = 0; i < ci_ext.num_ports; i++) {
      if (ci_ext.connections[i].port) {
        in = out = ci_ext.connections[i].port;
      } else {
        out = ci_ext.connections[i].end_point_info >> 4;
        in = ci_ext.connections[i].end_point_info & 0xf;
      }

      IO_TRACE(2) {
        char buf[5] = "----";

        *((int *) buf) = ci_ext.connections[i].port_function_id;
        fprintf(stderr,
                "ConnectionInfo_ext: entry %d function %s on port %d/%d\n",
                i, buf, in, out);

      }

/* XXX - byte order for sync */
      if (ci_ext.connections[i].port_function_id == 'cnys' ||
          ci_ext.connections[i].port_function_id == 'sync') {
        u->hotsync_ep_in = in;
        u->hotsync_ep_out = out;
        return 0;
      }
    }
  }
  return -1;
}

static int get_hotsync_ep(struct usb_data *u)
{
  int i;
  struct visor_connection_info ci;

  bzero((void *) &ci, sizeof(ci));
  if (usb_control_msg(u->dev, USB_TYPE_VENDOR | USB_RECIP_ENDPOINT | 0x80,
                      VISOR_GET_CONNECTION_INFORMATION, 0, 0, (char *) &ci,
                      6, 500) < 0) {
    perror(_("usb_control_msg(VISOR_GET_CONNECTION_INFORMATION) failed"));
    return -1;
  }

  /*
   *  Now search the list of functions supported over the USB interface
   *  for the endpoint associated with the HotSync function.  So far,
   *  this has seen to "always" be on endpoint 2, but this might 
   *  change and we use this binding mechanism to discover where it
   *  lives now.  Sort of like the portmap(8) daemon..
   *
   *  Also, beware:  as this is "raw" USB function, the result
   *  we get is in USB-specific byte order.  This happens to be
   *  little endian, but we should use the accessor macros to ensure
   *  the code continues to run on big endian CPUs too.
   */

  /* XXX Fix endian for ci.numPorts */
  for (i = 0; i < ci.num_ports; i++) {
    IO_TRACE(2)
        fprintf(stderr,
                "ConnectionInfo: entry %d function %s on port %d\n",
                i, (ci.connections[i].port_function_id <= 5)
                ? hs_usb_functions[ci.connections[i].port_function_id]
                : "unknown", ci.connections[i].port);

    if (ci.connections[i].port_function_id == VISOR_FUNCTION_HOTSYNC) {
      u->hotsync_ep_in = u->hotsync_ep_out = ci.connections[i].port;
      return 0;
    }
  }

  return -1;
}

int pconn_libusb_open(PConnection * pconn,
                      const char *device, const pconn_proto_t protocol)
{
  struct usb_data *u;
  int i, ret;
  unsigned char usbresponse[50];

  struct usb_bus *bus;
  struct usb_device *dev;

  /* Set the methods used by the USB connection */
  pconn->io_bind = &libusb_bind;
  pconn->io_read = &libusb_read;
  pconn->io_write = &libusb_write;
  pconn->io_connect = &libusb_connect;
  pconn->io_accept = &libusb_accept;
  pconn->io_close = &pconn_libusb_close;
  pconn->io_select = &libusb_select;
  pconn->io_drain = &libusb_drain;

  u = pconn->io_private = malloc(sizeof(struct usb_data));
  if (!u)
    return -1;

  bzero((void *) pconn->io_private, sizeof(struct usb_data));

  /*
   *  Prompt for the Hot Sync button now, as the USB bus
   *  enumerator won't create the underlying device we want to
   *  open until that happens.  The act of starting the hot sync
   *  operation on the PDA logically plugs it into the USB
   *  hub port, where it's noticed and enumerated.
   */
  if (pconn->flags & PCONNFL_PROMPT)
    printf(_("Please press the HotSync button.\n"));

  /*
   *  We've got to loop trying to open the USB device since
   *  you'll get an ENXIO until the device has been inserted
   *  on the USB bus.
   */

  usb_init();

  dev = NULL;
/* XXX - why stop after 100 seconds? */
  for (i = 0; i < 100; i++) {
    struct usb_bus *busses;

/* XXX - do one find_buses for 3 or 5 find_devices??? */
    usb_find_busses();
/* XXX - use result - if zero nothing changed */
    usb_find_devices();

    busses = usb_get_busses();

    for (bus = busses; bus; bus = bus->next) {

      IO_TRACE(10) {
        fprintf(stderr, "bus %x\n", (int) bus);
      }
      for (dev = bus->devices; dev; dev = dev->next) {
        /* XXX We should evaluate the devices, not the vendors */
/* XXX - find changes and fork if new device added */

        IO_TRACE(10) {
          fprintf(stderr,
                  "device vendor %04x product %04x\n",
                  dev->descriptor.idVendor, dev->descriptor.idProduct);

        }

        if (dev->descriptor.idVendor == HANDSPRING_VENDOR_ID)
          goto found;

        if (dev->descriptor.idVendor == PALM_VENDOR_ID)
          goto found;

        if (dev->descriptor.idVendor == SONY_VENDOR_ID)
          goto found;
      }
    }

    if (dev)
      break;

    IO_TRACE(2) {
      fprintf(stderr, ".");
    }

/* XXX - make delay configurable via coldsync.conf */
    sleep(1);
  }

found:
  /*
   *  Open the control endpoint of the USB device.  We'll use this
   *  to figure out if the device in question is the one we are
   *  interested in and understand, and then to configure it in
   *  preparation of doing I/O for the actual hot sync operation.
   */
  if (!dev || !(u->dev = usb_open(dev))) {
    fprintf(stderr, _("%s: Can't open USB device.\n"),
            "pconn_libusb_open");
    if (dev)
      perror("open");

    free(u);
    u = pconn->io_private = NULL;

    return -1;
  }

  IO_TRACE(1) {
    char buf[255] = "none";

    if (dev->descriptor.iSerialNumber)
      usb_get_string_simple(u->dev, dev->descriptor.iSerialNumber, buf,
                            255);

    fprintf(stderr,
            "USB device information: vendor %04x product %04x serial %s\n",
            dev->descriptor.idVendor, dev->descriptor.idProduct, buf);
  }

  /* Choice the default protocol */

  if (protocol == PCONN_STACK_DEFAULT) {
    if (dev->descriptor.idVendor == HANDSPRING_VENDOR_ID)
      /* XXX - what about treo? */
      pconn->protocol = PCONN_STACK_FULL;
    else
      pconn->protocol = PCONN_STACK_NET;
  } else
    pconn->protocol = protocol;

  /* Initialize the various protocols that the serial connection will
   * use.
   */
  switch (pconn->protocol) {
  case PCONN_STACK_FULL:
    /* Initialize the SLP part of the PConnection */
    if (slp_init(pconn) < 0) {
      return -1;
    }

    /* Initialize the PADP part of the PConnection */
    if (padp_init(pconn) < 0) {
      padp_tini(pconn);
      slp_tini(pconn);
      return -1;
    }

    /* Initialize the DLP part of the PConnection */
    if (dlp_init(pconn) < 0) {
      dlp_tini(pconn);
      padp_tini(pconn);
      slp_tini(pconn);
      return -1;
    }
    break;

  case PCONN_STACK_SIMPLE:     /* Fall through */
  case PCONN_STACK_NET:
    /* Initialize the DLP part of the PConnection */
    if (dlp_init(pconn) < 0) {
      dlp_tini(pconn);
      return -1;
    }

    /* Initialize the NetSync part of the PConnnection */
    if (netsync_init(pconn) < 0) {
      dlp_tini(pconn);
      netsync_tini(pconn);
      return -1;
    }
    break;

  case PCONN_STACK_NONE:
  case PCONN_STACK_DEFAULT:
    /* XXX - Error */
  default:
    /* XXX - Indicate error: unsupported protocol */
    return -1;
  }

  /*
   *  Happily, all of the multibyte values in the struct usb_device_info
   *  are in host byte order, and don't need to be converted.
   */

  /*
   *  Eventually, it might be necessary to split out the following
   *  code should another Palm device with a USB peripheral interface
   *  need to be supported in a different way.  Hopefully, they will
   *  simply choose to inherit this existing interface rather then
   *  inventing Yet Another exquisitely round wheel of their own.
   */

  /*
   *  Ensure that the device is set to the default configuration.
   *  For Visors seen so far, the default is the only one they
   *  support.
   */

/* XXX - after set_configuration on sony devices several (3) first usb packets
   from handheld are dropped (libusb bug?). but all tested devices (sony and
   palm) work well without this call.
   
   
  if (usb_set_configuration(u->dev, 1) < 0)
    perror("warning: usb_set_configuration failed");
*/

  if (usb_claim_interface(u->dev, 0) < 0)
    perror("warning: usb_claim_interface failed");

  /* XXX
   *  Now, ask the device (which we believe to be a Handspring Visor)
   *  about the various USB endpoints which we can make a connection 
   *  to.  Obviously, here we're looking for the endpoint associated 
   *  with the Hotsync running on the other end.  This has been observed
   *  to be endpoint "2", but this is not statically defined, and might
   *  change based on what other applications are running and perhaps
   *  on future hardware platforms.
   */

  if (get_hotsync_ep_ext(u))
    if (get_hotsync_ep(u)) {
      fprintf(stderr,
              _("%s: Could not find HotSync endpoint on your PDA.\n"),
              "PConnection_usb");

      usb_close(u->dev);

      free((void *) u);
      u = pconn->io_private = NULL;

      dlp_tini(pconn);
      padp_tini(pconn);
      slp_tini(pconn);
      return -1;
    }

/*
u->hotsync_ep_in = u->hotsync_ep_out = 2;
*/

  ret =
      usb_control_msg(u->dev, USB_TYPE_VENDOR | USB_RECIP_ENDPOINT | 0x80,
                      VISOR_REQUEST_BYTES_AVAILABLE, 0, 5,
                      (char *) usbresponse, 2, 500);

  if (ret < 0)
    perror(_("usb_control_msg(VISOR_REQUEST_BYTES_AVAILABLE) failed"));

  IO_TRACE(2) {
    if (ret > 0) {
      for (i = 0; i < ret; i++)
        fprintf(stderr, " 0x%02x", usbresponse[i]);

      fprintf(stderr, "\n");
    }
  }

  if ((usbresponse[0] | (usbresponse[1] << 8)) != 1) {
    fprintf(stderr,
            _("%s: unexpected response %d to "
              "GetBytesAvailable.\n"),
            "PConnection_usb", usbresponse[0] | (usbresponse[1] << 8));
  }

  /* Make it so */
  return 1;
}

#endif                          /* WITH_LIBUSB */
